from typing import Callable

import structlog

__all__ = [
    "FilterKeys",
    "FilterMethods",
    "exclude_keys",
    "exclude_methods",
    "only_keys",
    "only_methods",
]


class FilterMethods:
    """
    A processor that filters log events based on the name of the method that
    was called on the BoundLogger.
    """

    def __init__(
        self,
        predicate: Callable[[str], bool],
    ):
        self.predicate = predicate

    def __call__(
        self,
        logger: structlog.types.WrappedLogger,
        method_name: str,
        event_dict: structlog.types.EventDict,
    ) -> structlog.types.EventDict:
        if self.predicate(method_name):
            return event_dict
        else:
            raise structlog.DropEvent


def only_methods(*names: str) -> FilterMethods:
    """
    A specialisation of FilterMethods that only passes log events through if
    the method name is contained in 'names'.
    """
    return FilterMethods(lambda name: name in names)


def exclude_methods(*names: str) -> FilterMethods:
    """
    A specialisation of FilterMethods that drops log events if the method name
    is contained in 'names'.
    """
    return FilterMethods(lambda name: name not in names)


class FilterKeys:
    """
    A processor that filters the keys included in the event dictionary.
    """

    def __init__(
        self,
        predicate: Callable[[str], bool],
    ):
        self.predicate = predicate

    def __call__(
        self,
        logger: structlog.types.WrappedLogger,
        method_name: str,
        event_dict: structlog.types.EventDict,
    ) -> structlog.types.EventDict:
        for key in set(event_dict.keys()):
            if not self.predicate(key):
                del event_dict[key]
        return event_dict


def only_keys(*keys: str) -> FilterKeys:
    """
    A specialisation of FilterKeys that only keeps the keys contained in 'keys'.
    """
    return FilterKeys(lambda key: key in keys)


def exclude_keys(*keys: str) -> FilterKeys:
    """
    A specialisation of FilterKeys that removes the keys contained in 'keys'.
    """
    return FilterKeys(lambda key: key not in keys)



from typing import Any, Callable, List, Sequence, Type, cast

import structlog
from dataclasses import dataclass, field

__all__ = ["TeeOutput", "TeeLoggerFactory", "ConfigurationError"]


@dataclass
class TeeOutput:
    """
    A specification of a destination that a TeeLoggerFactory should send
    log events to. A chain of destination-specific processors can be
    applied before the log event ultimately makes it way to a logger created
    from the underlying factory.
    """

    logger_factory: Callable[..., structlog.types.WrappedLogger]
    processors: Sequence[structlog.types.Processor] = field(default_factory=list)

    def _construct_bound_logger(self, *args: Any) -> structlog.BoundLogger:
        return cast(
            structlog.BoundLogger,
            structlog.wrap_logger(
                logger=self.logger_factory(*args),
                processors=self.processors,
            ),
        )


class TeeLogger:
    def __init__(self, destinations: List[structlog.BoundLogger]):
        self.destinations = destinations

    def __getattr__(self, method_name: str) -> Any:
        def f(*args_from_renderer: Any, **event_dict: Any) -> None:
            if len(args_from_renderer):
                raise ConfigurationError(
                    "TeeLoggerFactory cannot operate on events that have "
                    "already been processed by a Renderer. Please place "
                    "the Renderer in the per-output processor lists instead "
                    "of the top-level list passed to structlog.configure."
                )
            for destination in self.destinations:
                getattr(destination, method_name)(**event_dict.copy())

        return f


_ensure_logger_implements_protocol: Type[structlog.types.WrappedLogger] = TeeLogger


class TeeLoggerFactory:
    """
    A logger factory that duplicates events to multiple destinations.

    For example, the following code sets up structlog to send events to both
    the console and a JSON file:

        import structlog, structlog_overtime, sys

        structlog.configure(
          # The default list of processors contains a ConsoleRenderer, but
          # TeeLoggerFactory cannot operate on rendered events so we must
          # remove it.
          processors=[
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.format_exc_info,
            structlog.processors.TimeStamper(),
          ],

          logger_factory=structlog_overtime.TeeLoggerFactory(
            structlog_overtime.TeeOutput(
              processors=[structlog.dev.ConsoleRenderer(colors=sys.stderr.isatty())],
              logger_factory=structlog.PrintLoggerFactory(sys.stderr),
            ),
            structlog_overtime.TeeOutput(
              processors=[structlog.processors.JSONRenderer()],
              logger_factory=structlog.PrintLoggerFactory(open("test.log", "a")),
            ),
          ),
        )
    """

    def __init__(self, *outputs: TeeOutput):
        self.outputs = outputs

    def __call__(self, *args: Any) -> TeeLogger:
        return TeeLogger(
            [output._construct_bound_logger(*args) for output in self.outputs]
        )


_ensure_logger_factory_implements_protocol: Type[
    Callable[..., structlog.types.WrappedLogger]
] = TeeLoggerFactory


class ConfigurationError(Exception):
    pass



from typing import Any, TypeVar, cast

import structlog

__all__ = ["bind"]

BL = TypeVar("BL", bound=structlog.types.BindableLogger)


def bind(logger: BL, **new_values: Any) -> BL:
    """
    A plain wrapper around logger.bind(**new_values) that preserves the type
    annotation of the logger. Does not check at runtime that the return value
    is indeed the assumed type.

    Hopefully this will no longer be needed in a future version of structlog
    (see https://www.structlog.org/en/stable/types.html).
    """
    return cast(BL, logger.bind(**new_values))



import structlog

__all__ = ["noop"]


def noop(
    logger: structlog.types.WrappedLogger,
    method_name: str,
    event_dict: structlog.types.EventDict,
) -> structlog.types.EventDict:
    """
    No operation (a processor that returns the event dictionary unchanged).
    """
    return event_dict



import functools
from typing import Any, Callable, List, Tuple, Type

import structlog
from dataclasses import dataclass, field

__all__ = ["MockLogRecord", "MockLoggerFactory"]


@dataclass(frozen=True)
class MockLogRecord:
    """
    A record of a log event observed by a MockLoggerFactory.

    For example, this call:

        structlog.get_logger("my_logger_name").info("hello world", a=1, b=2)

    would typically result in the following MockLogRecord:

        MockLogRecord(
            method_name = "info",
            event = {"event": "hello world", "a": 1, "b": 2},
            get_logger_args = ("my_logger_name",),
        )

    The underlying_logger_args field is relevent if a renderer-style
    processor is applied to the log event before it is passed to the
    MockLoggerFactory. For example, this snippet:

        structlog.configure(
            processors=[structlog.dev.ConsoleRenderer()],
            logger_factory=MockLoggerFactory(),
        )
        structlog.get_logger("my_logger_name").info("hello world", a=1, b=2)

    would result in the following MockLogRecord:

        MockLogRecord(
            method_name = "info",
            get_logger_args = ("my_logger_name",),
            underlying_logger_args = ("hello world                    a=1 b=2\n",),
        )
    """

    method_name: str
    event: structlog.types.EventDict = field(default_factory=dict)
    get_logger_args: Tuple[Any, ...] = ()
    underlying_logger_args: Tuple[Any, ...] = ()


@dataclass
class MockLogger:
    records: List[MockLogRecord]
    get_logger_args: Tuple[Any, ...]

    def __getattr__(self, method_name: str) -> Any:
        return functools.partial(self._emit, method_name)

    def _emit(self, method_name: str, *underlying_logger_args: Any, **event: Any) -> None:
        self.records.append(
            MockLogRecord(
                method_name=method_name,
                event=event,
                get_logger_args=self.get_logger_args,
                underlying_logger_args=underlying_logger_args,
            )
        )


@dataclass
class MockLoggerFactory:
    """
    A logger factory that accumulates log events in a list. Useful for tests.
    """

    records: List[MockLogRecord] = field(default_factory=list)

    def __call__(self, *args: Any) -> structlog.types.WrappedLogger:
        return MockLogger(records=self.records, get_logger_args=args)


_ensure_logger_factory_implements_protocol: Type[
    Callable[..., structlog.types.WrappedLogger]
] = MockLoggerFactory



from .bind import bind
from .filter import (
    FilterKeys,
    FilterMethods,
    exclude_keys,
    exclude_methods,
    only_keys,
    only_methods,
)
from .mock import MockLogger, MockLoggerFactory, MockLogRecord
from .noop import noop
from .tee import ConfigurationError, TeeLoggerFactory, TeeOutput
from .timestamper import TimezoneAwareTimeStamper

__all__ = [
    "bind",
    "ConfigurationError",
    "FilterKeys",
    "FilterMethods",
    "MockLogRecord",
    "MockLogger",
    "MockLoggerFactory",
    "TeeLoggerFactory",
    "TeeOutput",
    "TimezoneAwareTimeStamper",
    "exclude_keys",
    "exclude_methods",
    "noop",
    "only_keys",
    "only_methods",
]



import datetime
from typing import Callable

import structlog

__all__ = ["TimezoneAwareTimeStamper"]


def _default_now() -> datetime.datetime:
    return datetime.datetime.now().astimezone()  # pragma: no cover


class TimezoneAwareTimeStamper:
    """
    A workaround for the fact that structlog.processors.TimeStamper only
    supports UTC or naive (timezone-less) timestamps. Use this if you
    want to make the timezone explicit in the formatted log message.
    Full datetime.strftime syntax is supported.
    """

    def __init__(
        self,
        *,
        format: str = "%Y-%m-%dT%H:%M:%S%z",
        key: str = "timestamp",
        now: Callable[[], datetime.datetime] = _default_now,
    ):
        self.format = format
        self.key = key
        self.now = now

    def __call__(
        self,
        logger: structlog.types.WrappedLogger,
        method_name: str,
        event_dict: structlog.types.EventDict,
    ) -> structlog.types.EventDict:
        event_dict[self.key] = self.now().strftime(self.format)
        return event_dict



from pathlib import Path

from setuptools import find_packages, setup

setup(
    name="structlog-overtime",
    version="1.0.1",
    description="Miscellaneous utilities for structlog",
    long_description=Path(__file__).resolve().parent.joinpath("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Kier Davis",
    author_email="me@kierdavis.com",
    url="https://github.com/kierdavis/structlog-overtime",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Typing :: Typed",
    ],
    packages=find_packages(exclude=["tests", "stubs"]),
    include_package_data=True,
    python_requires=">= 3.6",
    install_requires=[
        "dataclasses; python_version < '3.7'",
        "structlog >= 20.2.0",
        # ...
    ],
    extras_require={
        "dev": [
            "black",
            "carthorse",
            "flake8",
            "isort",
            "mypy",
            "pytest",
            "pytest-cov",
            "setuptools>=38.6.0",  # for long_description_content_type support
            "testfixtures",
            "twine",
            "wheel",
        ],
    },
    package_data={
        "structlog_overtime": ["py.typed"],
    },
)
